const express = require("express");
const router = express.Router();
const RegisterRequestSchema = require("./../../model/patient/registration");
require("dotenv").config();
const StellarSDK = require("stellar-sdk");
const DecryptWithServerKey = require("./../../services/decryptionservice");
const EncryptionService = require("./../../services/encryptionservice");
const AccountCreateCollection = require("../../model/stellar/accountCreate");
const UserSchema = require("../../model/patient/patient");
let timestamp;

router.post("/", async (req, res) => {
  //decrypt user request details
  const decryptedEmail = DecryptWithServerKey(req.body.email);
  const decryptedPassword = DecryptWithServerKey(req.body.password);
  const decryptedUserName = DecryptWithServerKey(req.body.email);

  //catching the request
  const request = new RegisterRequestSchema({
    email: req.body.email,
    username: req.body.username,
    password: req.body.password,
  });

  console.log(request);

  //get the server key pair
  const serverSourceKeyPair = StellarSDK.Keypair.fromSecret(
    process.env.STELLARSEED
  );
  const serverPubKey = serverSourceKeyPair.publicKey();
  const server = new StellarSDK.Server(process.env.STELLARTESTNET);

  //load server account
  const account = await server.loadAccount(serverPubKey);
  const fee = await server.fetchBaseFee();

  //generate random keypair
  const userKeyPair = StellarSDK.Keypair.random();

  //build the account funding transaction
  const accCreateTransaction = new StellarSDK.TransactionBuilder(account, {
    fee,
    networkPassphrase: StellarSDK.Networks.TESTNET,
  })
    .addMemo(StellarSDK.Memo.text("Create account"))
    .addOperation(
      StellarSDK.Operation.createAccount({
        destination: userKeyPair.publicKey(),
        startingBalance: "10",
      })
    )
    .setTimeout(30)
    .build();

  //sign the transaction
  accCreateTransaction.sign(serverSourceKeyPair);

  //send account creation transaction to BC
  try {
    const accCreateTxnResult = await server.submitTransaction(
      accCreateTransaction
    );
    console.log("New user account created: HASH : ", accCreateTxnResult.hash);
    console.log(
      "View account creation transaction at : https://horizon-testnet.stellar.org/transactions/" +
        accCreateTxnResult.hash
    );
    timestamp = new Date();
    //create the account creation details collection object
    const createAccountObj = new AccountCreateCollection({
      publickey: userKeyPair.publicKey(),
      amount: "10XLM",
      timestamp: timestamp,
      transactionhash: accCreateTxnResult.hash,
    });

    //save the account creation details to the database
    try {
      const saveInCollection = await createAccountObj.save();
      console.log(
        "Account creation details saved in the database " + saveInCollection
      );

      //encrypt user entered details with stellar seed
      const encEmail = EncryptionService.EncryptWithStellarSeed(
        request.email,
        userKeyPair.secret()
      );
      const encUsername = EncryptionService.EncryptWithStellarSeed(
        request.username,
        userKeyPair.secret()
      );
      const encPassword = EncryptionService.EncryptWithStellarSeed(
        request.password,
        userKeyPair.secret()
      );

      //get the user keypair
      const userSourceKeyPair = StellarSDK.Keypair.fromSecret(
        userKeyPair.secret()
      );
      const userSourcePublicKey = userSourceKeyPair.publicKey();

      //load user account
      const userAccount = await server.loadAccount(userSourcePublicKey);

      //build genesis transaction
      const genesisTransaction = new StellarSDK.TransactionBuilder(
        userAccount,
        {
          fee,
          networkPassphrase: StellarSDK.Networks.TESTNET,
        }
      )
        .addMemo(StellarSDK.Memo.text("Genesis"))
        .addOperation(
          StellarSDK.Operation.manageData({
            name: "Email Address",
            value: encEmail,
          })
        )
        .addOperation(
          StellarSDK.Operation.manageData({
            name: "Username",
            value: encUsername,
          })
        )
        .addOperation(
          StellarSDK.Operation.manageData({
            name: "Previous",
            value: "",
          })
        )
        .setTimeout(30)
        .build();

      //sign the genesis transaction
      genesisTransaction.sign(userSourceKeyPair);

      //send the genesis transaction
      try {
        const genesisTxnResult = await server.submitTransaction(
          genesisTransaction
        );
        console.log(
          "Genesis transaction submitted: HASH : ",
          genesisTxnResult.hash
        );
        console.log(
          "View genesis transaction at : https://horizon-testnet.stellar.org/transactions/" +
            genesisTxnResult.hash
        );
        timestamp = new Date();
        console.log(new Date());
        //create the user collection object to be added to the DB
        const userCollectionObject = new UserSchema({
          username: req.body.username,
          password: req.body.password,
          email: req.body.email,
          publickey: userKeyPair.publicKey(),
          genesishash: genesisTxnResult.hash,
          currenthash: genesisTxnResult.hash,
          timestamp: timestamp,
        });

        //save the user collection object to the database
        try {
          const saveUserInCollection = await userCollectionObject.save();
          console.log("User details added to the DB " + saveUserInCollection);
          console.log("Public key: " + userKeyPair.publicKey());
          console.log("Secret: " + userKeyPair.secret());
          res.status(200).json({
            message: "User details added to the blockchain",
            hash: genesisTxnResult.hash,
            pubkey: EncryptionService.EncryptWithServerKey(
              userKeyPair.publicKey()
            ),
            seed: EncryptionService.EncryptWithServerKey(userKeyPair.secret()),
          });
          console.log(
            "/patient/register - register patient details successful"
          );
        } catch (error) {
          console.log(
            "Error when adding user details to the blockchain : " + error
          );
          res.status(500).json({
            message:
              "Error when adding user details to the blockchain : " + error,
          });
          console.log(
            "/patient/register - register patient details unsuccessful"
          );
        }
      } catch (error) {
        console.log(
          "Error when sending genesis transaction to blockchain: " + error
        );
        res.status(500).json({
          message:
            "Error when sending genesis transaction to blockchain: " + error,
        });
        console.log(
          "/patient/register - register patient details unsuccessful"
        );
      }
    } catch (error) {
      console.log(
        "Error when sending account creation transaction details to DB : " +
          error
      );
      res.status(500).json({
        message:
          "Error when sending account creation transaction details to DB : " +
          error,
      });
      console.log("/patient/register - register patient details unsuccessful");
    }
  } catch (error) {
    console.log(
      "Error when sending account creation transaction to blockchain: " + error
    );
    res.status(500).json({
      message:
        "Error when sending account creation transaction to blockchain: " +
        error,
    });
    console.log("/patient/register - register patient details unsuccessful");
  }
});

module.exports = router;
